# w.c.s. - web application for online forms
# Copyright (C) 2005-2018  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

from quixote import redirect
from quixote.directory import Directory
from quixote.errors import PublishError

from wcs.carddef import CardDef
from wcs.formdef import FormDef
from wcs.forms.common import FormTemplateMixin
from wcs.wf.jump import jump_and_perform

from ..qommon import _, errors, misc, template, tokens
from ..qommon.form import Form


class MissingOrExpiredToken(PublishError):
    status_code = 404
    title = _('Error')
    description = _('This action link has already been used or has expired.')


class MissingFormdata(PublishError):
    status_code = 404
    title = _('Error')
    description = _('This action link is no longer valid as the attached form has been removed.')


class ActionsDirectory(Directory):
    def _q_lookup(self, component):
        try:
            token = tokens.Token.get(component)
        except KeyError:
            raise MissingOrExpiredToken()
        if token.type != 'action':
            raise errors.TraversalError()
        return ActionDirectory(token)


class ActionDirectory(Directory, FormTemplateMixin):
    _q_exports = ['']
    templates = ['wcs/action.html']

    def __init__(self, token):
        self.token = token
        formdef_type = self.token.context.get('form_type', 'formdef')
        if formdef_type == 'carddef':
            formdef_class = CardDef
        elif formdef_type == 'formdef':
            formdef_class = FormDef
        else:
            raise errors.TraversalError()

        self.formdef = formdef_class.get_by_urlname(self.token.context['form_slug'])
        try:
            self.formdata = self.formdef.data_class().get(self.token.context['form_number_raw'])
        except KeyError:
            raise MissingFormdata()
        self.action = None
        status = self.formdata.get_status()
        for item in status.items:
            if getattr(item, 'identifier', None) == self.token.context['action_id']:
                self.action = item
                break
        else:
            raise MissingOrExpiredToken()

    def _q_index(self):
        template.html_top(title=self.formdef.name)
        form = Form()
        form.add_submit('submit', misc.site_encode(self.token.context['label']))
        if form.is_submitted() and not form.has_errors():
            return self.submit()
        context = {
            'view': self,
            'form': form,
            'html_form': form,
            'message': self.token.context.get('message'),
        }
        return template.QommonTemplateResponse(
            templates=list(self.get_formdef_template_variants(self.templates)), context=context
        )

    def submit(self):
        url = jump_and_perform(self.formdata, self.action, event='email-button')
        done_message = self.token.context.get('done_message')
        self.token.remove_self()
        if url:
            return redirect(url)
        context = {
            'view': self,
            'done': True,
            'done_message': done_message,
        }
        return template.QommonTemplateResponse(
            templates=list(self.get_formdef_template_variants(self.templates)), context=context
        )
