# w.c.s. - web application for online forms
# Copyright (C) 2005-2020  Entr'ouvert
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, see <http://www.gnu.org/licenses/>.

from collections import defaultdict

from quixote import get_publisher

from wcs.categories import MailTemplateCategory
from wcs.qommon import _, get_logger
from wcs.qommon.form import OptGroup
from wcs.qommon.xml_storage import XmlStorableObject


class MailTemplate(XmlStorableObject):
    _names = 'mail-templates'
    xml_root_node = 'mail-template'
    backoffice_class = 'wcs.admin.mail_templates.MailTemplatePage'
    verbose_name = _('Mail template')
    verbose_name_plural = _('Mail templates')

    name = None
    slug = None
    description = None
    subject = None
    body = None
    attachments = []
    category_id = None

    # declarations for serialization
    XML_NODES = [
        ('name', 'str'),
        ('slug', 'str'),
        ('description', 'str'),
        ('subject', 'str'),
        ('body', 'str'),
        ('attachments', 'str_list'),
    ]

    def __init__(self, name=None):
        XmlStorableObject.__init__(self)
        self.name = name

    @property
    def category(self):
        return MailTemplateCategory.get(self.category_id, ignore_errors=True)

    @category.setter
    def category(self, category):
        if category:
            self.category_id = category.id
        elif self.category_id:
            self.category_id = None

    def get_admin_url(self):
        base_url = get_publisher().get_backoffice_url()
        return '%s/workflows/mail-templates/%s/' % (base_url, self.id)

    def store(self, comment=None, *args, **kwargs):
        assert not self.is_readonly()
        if self.slug is None:
            # set slug if it's not yet there
            self.slug = self.get_new_slug()
        super().store(*args, **kwargs)
        if get_publisher().snapshot_class:
            get_publisher().snapshot_class.snap(instance=self, comment=comment)

    def get_places_of_use(self):
        from wcs.workflows import Workflow

        for workflow in Workflow.select(ignore_errors=True, ignore_migration=True):
            for item in workflow.get_all_items():
                if item.key != 'sendmail':
                    continue
                if item.mail_template == self.slug:
                    yield workflow
                    break

    def is_in_use(self):
        return any(self.get_places_of_use())

    @classmethod
    def get_as_options_list(cls):
        mail_templates_by_category_names = defaultdict(list)
        for mail_template in cls.select(order_by='name'):
            name = ''
            if mail_template.category:
                name = mail_template.category.name
            mail_templates_by_category_names[name].append(mail_template)
        category_names = list(mail_templates_by_category_names.keys())
        if len(category_names) == 1 and category_names[0] == '':
            # no category found
            return [(mt.slug, mt.name, mt.slug) for mt in mail_templates_by_category_names['']]
        options = []
        # sort categories
        category_names = sorted(category_names)
        # mail template without categories at the end
        if category_names[0] == '':
            category_names = category_names[1:] + ['']
        # group by category name
        for name in category_names:
            options.append(OptGroup(name or _('Without category')))
            options.extend([(mt.slug, mt.name, mt.slug) for mt in mail_templates_by_category_names[name]])
        return options

    @classmethod
    def get_by_slug(cls, slug, ignore_errors=True):
        mail_template = super().get_by_slug(slug, ignore_errors=ignore_errors)
        if mail_template is None:
            get_logger().warning("mail template '%s' does not exist" % slug)
        return mail_template

    def get_dependencies(self):
        yield self.category

    def export_to_xml(self, include_id=False):
        root = super().export_to_xml(include_id=include_id)
        MailTemplateCategory.object_category_xml_export(self, root, include_id=include_id)
        return root

    @classmethod
    def import_from_xml_tree(cls, tree, include_id=False, **kwargs):
        mail_template = super().import_from_xml_tree(tree, include_id=include_id, **kwargs)
        MailTemplateCategory.object_category_xml_import(mail_template, tree, include_id=include_id)
        return mail_template
